/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.solutions.identity.cards;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import org.apache.axiom.om.util.Base64;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Element;
import org.wso2.solutions.identity.IdentityConstants;
import org.wso2.solutions.identity.IdentityProviderConstants;
import org.wso2.solutions.identity.IdentityProviderException;
import org.wso2.solutions.identity.admin.ParameterAdmin;
import org.wso2.solutions.identity.admin.RegisteredInfoCardInfoAdmin;
import org.wso2.solutions.identity.admin.ReportAdmin;
import org.wso2.solutions.identity.cards.model.CardImage;
import org.wso2.solutions.identity.cards.model.CardModelException;
import org.wso2.solutions.identity.cards.model.Identity;
import org.wso2.solutions.identity.cards.model.InformationCard;
import org.wso2.solutions.identity.cards.model.InformationCardReference;
import org.wso2.solutions.identity.cards.model.Metadata;
import org.wso2.solutions.identity.cards.model.RequireAppliesTo;
import org.wso2.solutions.identity.cards.model.SelfIssuedCredential;
import org.wso2.solutions.identity.cards.model.SupportedClaimType;
import org.wso2.solutions.identity.cards.model.SupportedClaimTypeList;
import org.wso2.solutions.identity.cards.model.TokenService;
import org.wso2.solutions.identity.cards.model.TokenServiceList;
import org.wso2.solutions.identity.cards.model.UserCredential;
import org.wso2.solutions.identity.cards.model.UsernamePasswordCredential;
import org.wso2.solutions.identity.i18n.Messages;
import org.wso2.solutions.identity.persistence.IPPersistenceManager;
import org.wso2.solutions.identity.persistence.dataobject.ActionDO;
import org.wso2.solutions.identity.persistence.dataobject.ClaimDO;
import org.wso2.solutions.identity.persistence.dataobject.InfoCardDO;
import org.wso2.utils.ServerConfiguration;
import org.wso2.wsas.ServerConstants;

/**
 * Issues information cards according to the given configuration.
 */
public class CardIssuer {

    private static Log log = LogFactory.getLog(CardIssuer.class);

    private static Messages messages = Messages
            .getInstance(IdentityProviderConstants.RESOURCES);

    private static CardIssuerConfig issuerConfig = null;

    public CardIssuer() throws IdentityProviderException {
        issuerConfig = CardIssuerConfig.getInstance();
    }

    public Element issueCardForUsername(String username,
            boolean requireAppliesTo) throws IdentityProviderException {

        UsernamePasswordCredential passCred = new UsernamePasswordCredential();
        passCred.setUsername(username);
        UserCredential cred = new UserCredential(passCred);
        return issueCard(cred, requireAppliesTo, username);

    }

    public Element issueCardForSelfIssuedCard(String ppid,
            boolean requireAppliesTo) throws IdentityProviderException {
        Element cardElement = null;
        SelfIssuedCredential selfCred = new SelfIssuedCredential(ppid);
        RegisteredInfoCardInfoAdmin registerAdmin = new RegisteredInfoCardInfoAdmin();
        String primaryUserName = registerAdmin.extractPrimaryUserName(ppid);
        UserCredential cred = new UserCredential(selfCred);
        
        if(primaryUserName == null){
            cardElement = issueCard(cred, requireAppliesTo, ppid);
        }else{
            cardElement = issueCard(cred, requireAppliesTo, primaryUserName);
        }
        
        return cardElement;
    }

    private Element issueCard(UserCredential credential,
            boolean requireAppliesTo, String primaryUserId) throws IdentityProviderException {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();

        try {
            String storeFilePath = serverConfig
                    .getFirstProperty("Security.KeyStore.Location");
            FileInputStream is = new FileInputStream(storeFilePath);

            KeyStore store = KeyStore.getInstance(serverConfig
                    .getFirstProperty("Security.KeyStore.Type"));
            String passwd = serverConfig
                    .getFirstProperty("Security.KeyStore.Password");
            store.load(is, passwd.toCharArray());

            Generator gen = new Generator();

            gen.setSignatureAlgorithm(issuerConfig.getSigAlgo());
            String alias = serverConfig
                    .getFirstProperty("Security.KeyStore.KeyAlias");
            Certificate[] certs = store.getCertificateChain(alias);
            gen.setCertCain(certs);
            gen.setPrivateKey((PrivateKey) store.getKey(alias, serverConfig
                    .getFirstProperty("Security.KeyStore.KeyPassword")
                    .toCharArray()));

            Identity id = new Identity();
            id.setCertificate((X509Certificate) store.getCertificate(alias));

            InformationCard infoCard = getInfoCard(credential, id,
                    requireAppliesTo);

            storeCard(infoCard, primaryUserId);
            Element elem = gen.signCard(infoCard);

            String cardId = infoCard.getInformationCardReference().getCardId();
           
            ReportAdmin.record(primaryUserId,
                    ActionDO.ACTION_USER_DOWNLOAD_CARD, "CardId=" + cardId);

            return elem;
        } catch (CardModelException e) {
            throw new IdentityProviderException("cardModelError", e);
        } catch (Exception e) {
            throw new IdentityProviderException("keyStoreException",
                    new String[] { serverConfig
                            .getFirstProperty("Security.KeyStore.Location") },
                    e);
        }
    }

    private InformationCard getInfoCard(UserCredential credential, Identity id,
            boolean requireAppliesTo) throws CardModelException,
            IdentityProviderException {

        InformationCard card = new InformationCard();

        InformationCardReference ref = new InformationCardReference(
                "http://identity.wso2.org/" + UUIDGenerator.getUUID(), 1);
        card.setInformationCardReference(ref);
        card.setIssuer(issuerConfig.getIssuer());

        Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
        Date now = cal.getTime();
        long lifetime = issuerConfig.getValidPeriod() * 1000l * 60l * 60l * 24l;
        Date exp = new Date(now.getTime() + lifetime);

        if (log.isDebugEnabled()) {
            DateFormat zulu = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
            log.info(messages.getMessage("createdInfocardAt",
                    new String[] { zulu.format(now) }));
            log.info(messages.getMessage("infocardExpiresAt",
                    new String[] { zulu.format(exp) }));
        }

        card.setTimeIssued(now);
        card.setTimeExpires(exp);

        card.setCardName("WSO2 Managed Card");

        byte[] imgBytes = getCardImageBytes();
        CardImage img = new CardImage("image/jpeg", Base64.encode(imgBytes));
        card.setCardImage(img);

        ServerConfiguration config = ServerConfiguration.getInstance();
        String host = config.getFirstProperty("HostName");
        String httpsPort = config.getFirstProperty("Ports.HTTPS");

        EndpointReference stsEpr = null;
        Metadata mexEpr = null;

        String stsAddress = "https://" + host + ":" + httpsPort
                + "/wsas/services/";
        String mexAddress = "https://" + host + ":" + httpsPort
                + "/wsas/services/";

        if (issuerConfig.isUseSymmetricBinding()) { // TODO post beta
            if (credential.getCredential() instanceof UsernamePasswordCredential) {
                stsAddress += IdentityProviderConstants.SERVICE_NAME_STS_UT_SYMM;
                mexAddress += IdentityProviderConstants.SERVICE_NAME_MEX_UT_SYMM
                        + "/get";
            } else if (credential.getCredential() instanceof SelfIssuedCredential) {
                stsAddress += IdentityProviderConstants.SERVICE_NAME_STS_IC_SYMM;
                mexAddress += IdentityProviderConstants.SERVICE_NAME_MEX_IC_SYMM
                        + "/get";
            }
        } else {
            if (credential.getCredential() instanceof UsernamePasswordCredential) {
                stsAddress += IdentityProviderConstants.SERVICE_NAME_STS_UT;
                mexAddress += IdentityProviderConstants.SERVICE_NAME_MEX_UT
                        + "/get";
            } else if (credential.getCredential() instanceof SelfIssuedCredential) {
                stsAddress += IdentityProviderConstants.SERVICE_NAME_STS_IC;
                mexAddress += IdentityProviderConstants.SERVICE_NAME_MEX_IC
                        + "/get";
            }
        }

        stsEpr = new EndpointReference(stsAddress);

        mexEpr = new Metadata(mexAddress);

        stsEpr.addExtensibleElement(id.serialize());
        stsEpr.addMetaData(mexEpr.serialize());

        TokenService service = new TokenService(stsEpr, credential);
        TokenServiceList serviceList = new TokenServiceList();
        serviceList.addTokenService(service);

        card.setTokenServiceList(serviceList);

        card.setSupportedTokenTypeList(issuerConfig.getTokenTypeList());

        // Get the list of supported claims
        IPPersistenceManager db = IPPersistenceManager.getPersistanceManager();
        ClaimDO[] supportedClaims = db.getAllSupportedClaims();
        SupportedClaimTypeList claimTypeList = new SupportedClaimTypeList();
        for (int i = 0; i < supportedClaims.length; i++) {
            if (supportedClaims[i].isSupported()) {
                SupportedClaimType claim = new SupportedClaimType(
                        supportedClaims[i].getUri());
                claim.setDisplayTag(supportedClaims[i].getDisplayTag());
                claim.setDescription(supportedClaims[i].getDescription());
                claimTypeList.addSupportedClaimType(claim);
            }
        }

        card.setSupportedClaimTypeList(claimTypeList);
        final RequireAppliesTo appliesTo = new RequireAppliesTo();
        if (requireAppliesTo) {
            card.setRequireAppliesTo(appliesTo);
        } else {
            appliesTo.setOptional(true);
            card.setRequireAppliesTo(appliesTo);
        }
        return card;

    }

    private byte[] getCardImageBytes() throws CardModelException {
        try {
            ParameterAdmin admin = new ParameterAdmin();
            FileInputStream is = new FileInputStream(System
                    .getProperty(ServerConstants.WSO2WSAS_HOME)
                    + IdentityConstants.CARD_IMAGE_PATH);
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            byte[] data = new byte[1024];
            int length = 0;
            while (is.available() > 0) {
                length = is.read(data);
                baos.write(data, 0, length);
            }

            return baos.toByteArray();
        } catch (Exception e) {
            throw new CardModelException(e.getMessage(), e);
        }
    }

    private void storeCard(InformationCard card, String userId)
            throws IdentityProviderException {

        IPPersistenceManager dbman = IPPersistenceManager
                .getPersistanceManager();
        InfoCardDO cardDo = new InfoCardDO();
        cardDo.setCardId(card.getInformationCardReference().getCardId());
        cardDo.setDateIssued(card.getTimeIssued());
        cardDo.setUserId(userId);
        cardDo.setDateExpires(card.getTimeExpires());

        dbman.create(cardDo);
        log.info("Information card details stored for card id : "
                + card.getInformationCardReference().getCardId());
    }

}
