/*                                                                             
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */
package org.wso2.solutions.identity.sts;

import org.apache.axiom.om.OMElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.rahas.RahasData;
import org.apache.xml.security.utils.Base64;
import org.wso2.solutions.identity.IdentityConstants;
import org.wso2.solutions.identity.IdentityProviderConstants;
import org.wso2.solutions.identity.IdentityProviderException;

import javax.xml.namespace.QName;

import java.net.URI;
import java.net.URISyntaxException;
import java.security.MessageDigest;

public class IdentityProviderUtil {

    private static Log log = LogFactory.getLog(IdentityProviderUtil.class);
            
    private final static char[] ppidDisplayCharMap = new char[] { 'Q', 'L',
        '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E',
        'F', 'G', 'H', 'J', 'K', 'M', 'N', 'P', 'R', 'S', 'T', 'U', 'V',
        'W', 'X', 'Y', 'Z' };

    public static OMElement createRequestedDisplayToken(OMElement parent,
            IdentityProviderData data) {

        return createOMElement(parent, IdentityConstants.NS,
                IdentityProviderConstants.LocalNames.REQUESTED_DISPLAY_TOKEN,
                IdentityConstants.PREFIX);
    }

    public static OMElement createDisplayToken(OMElement parent,
            IdentityProviderData data) {

        OMElement displayElem = createOMElement(parent, IdentityConstants.NS,
                IdentityProviderConstants.LocalNames.DISPLAY_TOKEN,
                IdentityConstants.PREFIX);

        return displayElem;

    }

    public static OMElement createDisplayClaim(OMElement parent,
            String displayTag, String displayValue, String uri) {
        OMElement claimElem = createOMElement(parent, IdentityConstants.NS,
                IdentityProviderConstants.LocalNames.DISPLAY_CLAIM,
                IdentityConstants.PREFIX);

        claimElem.addAttribute("Uri", uri, null);

        OMElement tagElem = createOMElement(claimElem, IdentityConstants.NS,
                IdentityProviderConstants.LocalNames.DISPLAY_TAG,
                IdentityConstants.PREFIX);
        tagElem.setText(displayTag);

        OMElement valElem = createOMElement(claimElem, IdentityConstants.NS,
                IdentityProviderConstants.LocalNames.DISPLAY_VALUE,
                IdentityConstants.PREFIX);
        valElem.setText(displayValue);

        return claimElem;
    }

    private static OMElement createOMElement(OMElement parent, String ns,
            String ln, String prefix) {

        return parent.getOMFactory().createOMElement(new QName(ns, ln, prefix),
                parent);

    }
    
    public static String getPPIDDisplayValue(String value) throws IdentityProviderException {
        try {
            log.info("Generating display value of PPID : " + value);
            byte[] rawPpid =  Base64.decode(value);
            MessageDigest sha1 = MessageDigest.getInstance("SHA1");
            sha1.update(rawPpid);
            byte[] hashId = sha1.digest();
            char[] returnChars = new char[10];
            for (int i = 0; i < 10; i++) {
                int rawValue = (hashId[i] + 128) % 32;
                returnChars[i] = ppidDisplayCharMap[rawValue];
            }
            StringBuffer sb = new StringBuffer();
            sb.append(returnChars, 0, 3);
            sb.append("-");
            sb.append(returnChars, 3, 4);
            sb.append("-");
            sb.append(returnChars, 6, 3);
            return sb.toString();
        } catch (Exception e) {
            throw new IdentityProviderException("errorInGeneratingPPIDDisplayValue", new String[] {value}, e);
        }
    }

    /**
     * Obtain the applies to host name from the WS-Trust request.
     * @param data Data from WS-Trust request.
     * @return applies to host name if found.
     * @throws IdentityProviderException
     */
    public static String getAppliesToHostName(RahasData data)
            throws IdentityProviderException {
        // If there's no applies to then we don't have to encrypt
        if (data.getAppliesToEpr() == null) {
            return null;
        }

        String relyingPartyURI = data.getAppliesToAddress();

        if (relyingPartyURI == null) {
            // Addressing policy not used in the policy
            relyingPartyURI = data.getAppliesToEpr().getText();
            if (relyingPartyURI == null) {
                throw new IdentityProviderException(
                        "cannotFindRelyingParty");
            }
        }
        
        URI uri = null;
        try {
            //To get the host name extracted
            uri = new URI(relyingPartyURI);
        } catch (URISyntaxException e) {
            throw new IdentityProviderException("invalidUri",
                    new String[] { relyingPartyURI }, e);
        }
        return uri.getHost();
    }

}
