/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <string.h>
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#if 0
#ifndef XMLSEC_NO_XSLT
#include <libxslt/xslt.h>
#endif /* XMLSEC_NO_XSLT */
#endif
#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/xmlenc.h>
#include <xmlsec/xmldsig.h>
#include <xmlsec/crypto.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <xmlsec/x509.h>

#include "cspace_debug.h"
#include "process_request.h"
#include "process_request_defines.h"

/* "/saml:Assertion" */
#define XPATH_ASSERTION "/" SAML_PFX ":" ASSERTION

/* "/saml:Assertion/dsig:Signature" */
#define XPATH_SIGNATURE XPATH_ASSERTION "/" DSIG_PFX ":" \
                        SIGNATURE

/* "/saml:Assertion/dsig:Signature/dsig:KeyInfo" */
#define XPATH_KEY_INFO XPATH_SIGNATURE "/" DSIG_PFX ":" KEY_INFO

/* "/saml:Assertion/dsig:Signature/dsig:KeyInfo/dsig:X509Data" */
#define XPATH_X509_DATA XPATH_KEY_INFO "/" DSIG_PFX ":" X509_DATA 

/* "/enc:EncryptedData" */
#define XPATH_ENC_DATA "/" ENC_PFX ":" ENCRYPTED_DATA

/* "/saml:Assertion/saml:AttributeStatement/saml:Attribute" */
#define XPATH_ATTRIBUTE XPATH_ASSERTION "/" SAML_PFX ":" ATTRIBUTE_STATEMENT \
                        "/" SAML_PFX ":" ATTRIBUTE

#define XPATH_OBJ_SIZE(obj) ((obj) ? (((obj)->nodesetval) ? \
                                      ((obj)->nodesetval->nodeNr) : 0) : 0)

#define XPATH_OBJ_TO_NODE(obj, i) \
           ((xmlNode *) \
            (obj) ? (((obj)->nodesetval) ? \
                     ((obj)->nodesetval->nodeTab[(i)]) : NULL) : NULL)

static void *malloc_def(size_t size, void *cb_ctx);
static void free_def(void *ptr, void *cb_ctx);
static void logger_def(const char *msg, pc_log_level_t level, void *cb_ctx);
static int decrypt_token(xmlDoc *doc, process_context_t *ctx);
static int register_namespaces(xmlXPathContext *xpath_ctx);
static xmlSecKeysMngr *keys_mnrg_create_and_load_priv_key(process_context_t *ctx);
static xmlXPathObject *cspace_xpath_evaluate(xmlDoc *doc, xmlChar *xpath_exp);
static int verify_signature(xmlDoc *doc, process_context_t *ctx);
static int register_id(xmlDoc *doc, xmlNodePtr node, const xmlChar* idName);
static int extract_info(process_context_t *ctx, xmlDoc *doc);

int cspace_process_request(process_context_t *ctx, const char *req)
{
    int res = FAIL;
    xmlDoc *doc = NULL;
 
    /* Load XML document */
    doc = xmlReadMemory(req, (int)strlen(req), NULL, NULL, 0);
    
    if (!decrypt_token(doc, ctx)) {
        /*log*/
        goto done;
    }
    
    if (!ctx->set_header_fn) {
        /*log*/
        goto done;
    }
    
    if (!extract_info(ctx, doc)) {
        goto done;
    }

    /* verify signature */
    if (!verify_signature(doc, ctx)) {
        /*log*/
        goto done;
    }

    res = SUCC;

done:
    if (doc)
        xmlFreeDoc(doc);

    /*add the "cardspace_state" header*/
    if (res) {
        (*ctx->set_header_fn)(NULL, cspace_strdup(CARDSPACE_HEADER_STATE,
                                            ctx->allocator),
                              cspace_strdup(CARDSPACE_STATE_SUCCESS,
                                            ctx->allocator), 
                              ctx->header_container,
                              ctx->set_header_cb_ctx);
    } else {
        (*ctx->set_header_fn)(NULL, cspace_strdup(CARDSPACE_HEADER_STATE,
                                            ctx->allocator),
                              cspace_strdup(CARDSPACE_STATE_FAIL,
                                            ctx->allocator), 
                              ctx->header_container,
                              ctx->set_header_cb_ctx);
    }

    return res;
}

void cspace_process_context_set_logger(process_context_t *ctx,
                                       logger_cb_t logger, void *cb_ctx)
{
    ctx->logger_fn = logger;
    ctx->logger_cb_ctx = cb_ctx;
}

void cspace_process_context_set_header_callback_context(process_context_t *ctx,
                                                        void *cb_ctx)
{
    ctx->set_header_cb_ctx = cb_ctx;
}

/*caller must free the return xmlXPathObject*/
static xmlXPathObject *cspace_xpath_evaluate(xmlDoc *doc, xmlChar *xpath_exp)
{
    xmlXPathContext *xpath_ctx = NULL;
    
    xmlXPathObject *xpath_obj = NULL; 
    /*need to be init to NULL, it will
     returned even in the error path*/
    
    /* Create xpath evaluation context */
    xpath_ctx = xmlXPathNewContext(doc);
    if (!xpath_ctx) {
        goto done;
    }
    
    /* Register namespaces from list (if any) */
    if (!register_namespaces(xpath_ctx)) {
        goto done;
    }

    xpath_obj = xmlXPathEvalExpression(xpath_exp, xpath_ctx);

done:
    if (xpath_ctx)  
        xmlXPathFreeContext(xpath_ctx);

    return xpath_obj;
}



static int verify_signature(xmlDoc *doc, process_context_t *ctx)
{
    int res = FAIL;
    xmlSecKeysMngr *vfy_mngr = NULL;
    xmlXPathObject *xpath_obj = NULL;
    xmlSecDSigCtx *dsig_ctx = NULL;
    xmlNode *vfy_node = NULL;

    vfy_mngr = xmlSecKeysMngrCreate();
    if(vfy_mngr == NULL) {
        fprintf(stderr, "Error: failed to create keys manager.\n");
        goto done;
    }

    if(xmlSecCryptoAppDefaultKeysMngrInit(vfy_mngr) < 0) {
        fprintf(stderr, "Error: failed to initialize keys manager.\n");
        goto done;
    }
    
    if (ctx->ca_file) {
        if(xmlSecCryptoAppKeysMngrCertLoad(vfy_mngr, ctx->ca_file,
                                           xmlSecKeyDataFormatPem,
                                           xmlSecKeyDataTypeTrusted) < 0) {
            fprintf(stderr,"Error: failed to load pem certificate from \n" );
            goto done;
        }
    }

    /* Add the ID for non-standard reference
       refer to http://ml.osdir.com/text.xml.xmlsec/2003-12/msg00019.html */
    if (!register_id(doc, xmlDocGetRootElement(doc), BAD_CAST ASSERTION_ID)) {
        /*log*/
        goto done;
    }

    dsig_ctx = xmlSecDSigCtxCreate(vfy_mngr);
    if (!dsig_ctx) {
        goto done;
    }

    xpath_obj = cspace_xpath_evaluate(doc, BAD_CAST XPATH_SIGNATURE);
    if (!xpath_obj) {
        goto done;
    }

    vfy_node = XPATH_OBJ_TO_NODE(xpath_obj, 0);
    
    if (xmlSecDSigCtxVerify(dsig_ctx, vfy_node) < 0) {
        goto done;
    }

    if (dsig_ctx->status != xmlSecDSigStatusSucceeded) {
        goto done;
    }

    res = SUCC;
#ifdef CSPACE_DEBUG
    printf("\nVERIFICATION SUCCESS!\n");
#endif

done:
    if (xpath_obj)
        xmlXPathFreeObject(xpath_obj);

    if (dsig_ctx)
        xmlSecDSigCtxDestroy(dsig_ctx);

    if (vfy_mngr)
        xmlSecKeysMngrDestroy(vfy_mngr);
    
    /* TODO: Add the free code */

    return res;
}


static int extract_info(process_context_t *ctx, xmlDoc *doc)
{
    int res = FAIL;
    int i;
    
    xmlXPathObject *attrs_obj = NULL;
    xmlNode *attr_node = NULL;
    xmlNode *text_node = NULL;

    char *attr_name = NULL;
    char *attr_val = NULL;
    char *attr_ns = NULL;

    attrs_obj = cspace_xpath_evaluate(doc, BAD_CAST XPATH_ATTRIBUTE);
    if (!attrs_obj) {
        goto done;
    }
    
    for (i = 0; i < XPATH_OBJ_SIZE(attrs_obj); i++) {
        
        attr_node = XPATH_OBJ_TO_NODE(attrs_obj, i);
        attr_name = (char *) xmlGetProp(attr_node, BAD_CAST ATTRIBUTE_NAME);
        attr_ns = (char *) xmlGetProp(attr_node, BAD_CAST ATTRIBUTE_NAMESPACE);

        /*TODO: validate the node structure*/
        attr_node = attr_node->children;
        if (!attr_node) {
           continue;
        }

        text_node = attr_node->children;
        if (text_node && xmlNodeIsText(text_node)) {
            attr_val = cspace_strdup((char *) text_node->content,
                                     ctx->allocator);
        }
        
        (*ctx->set_header_fn)(attr_ns, attr_name, attr_val,
                              ctx->header_container, ctx->set_header_cb_ctx);
        
    }

    res = SUCC;

done:
    if (attrs_obj)
        xmlXPathFreeObject(attrs_obj);

    return res;
}

/* Add the ID for non-standard reference
   refer to http://ml.osdir.com/text.xml.xmlsec/2003-12/msg00019.html */
static int register_id(xmlDoc *doc, xmlNode *node, const xmlChar* idName)
{
    xmlAttr *attr;
    xmlAttr *tmp;
    xmlChar *name;
        
    /* find pointer to id attribute */
    attr = xmlHasProp(node, idName);
    if((attr == NULL) || (attr->children == NULL)) {
        return FAIL;
    }
    
    /* get the attribute (id) value */
    name = xmlNodeListGetString(doc, attr->children, 1);
    if(name == NULL) {
        return FAIL;    
    }
    
    /* check that we don't have that id already registered */
    tmp = xmlGetID(doc, name);
    if(tmp != NULL) {
        xmlFree(name);
        return FAIL;
    }
    
    /* finally register id */
    xmlAddID(NULL, doc, name, attr);

    /* and do not forget to cleanup */
    xmlFree(name);

    return SUCC;
}

static xmlSecKeysMngr *keys_mnrg_create_and_load_priv_key(process_context_t *ctx)
{
    xmlSecKeysMngr *mngr = NULL;
    xmlSecKey *key = NULL;

    /*load pem key*/
    mngr = xmlSecKeysMngrCreate();
    if (!mngr) {
        /*log error*/
        goto error_handler;
    }

    if (xmlSecCryptoAppDefaultKeysMngrInit(mngr) < 0) {
        /*log error*/
        goto error_handler;
    }

    key = xmlSecCryptoAppKeyLoad(ctx->key_file, xmlSecKeyDataFormatPem, 
                                 NULL/*password*/, NULL, NULL);
    if (!key) {
        /*log error*/
        goto error_handler;
    }

    if (xmlSecKeySetName(key, BAD_CAST ctx->key_file) < 0) {
        /*log error*/
        goto error_handler;
    }

    if (xmlSecCryptoAppDefaultKeysMngrAdoptKey(mngr, key) < 0) {
        /*log error*/
        goto error_handler;
    }

    /*load pem key successful -- now the key is free'd by the manager*/
    return mngr;

error_handler:
    if (mngr)
        xmlSecKeysMngrDestroy(mngr);

    return NULL;
}


static int decrypt_token(xmlDoc *doc, process_context_t *ctx)
{
    int res = FAIL;
    xmlNode *node = NULL;
    xmlSecEncCtx *enc_ctx = NULL;
    xmlSecKeysMngr *mngr = NULL;
    xmlSecKey *key = NULL;
    xmlXPathObject *xpath_obj = NULL;

/* this part is moved to keys_mnrg_create_and_load_priv_key() */
#if 0 /* this part is moved to */
    /* load the key and create keys manager */
    key = xmlSecCryptoAppKeyLoad(ctx->key_file, xmlSecKeyDataFormatPem, 
                                 NULL/*password*/, NULL, NULL);
    if (!key) {
        /*log error*/
        goto done;
    }

    if (xmlSecKeySetName(key, BAD_CAST ctx->key_file) < 0) {
        /*log error*/
        goto done;
    }
#endif

    mngr = keys_mnrg_create_and_load_priv_key(ctx);
    if (!mngr) {
        /*log*/
        goto done;
    }
    key = NULL; /* we no longer should worry about freeing the key
                   keys manager does it from now on */

    /* done with creating keys manager loaded with the key */


    /* find start node */
    /*node = xmlSecFindNode(xmlDocGetRootElement(doc), xmlSecNodeEncryptedData,
                          xmlSecEncNs);*/
    xpath_obj = cspace_xpath_evaluate(doc, BAD_CAST XPATH_ENC_DATA);
    if (!xpath_obj) {
        /*log error*/
        goto done;
    }

    node = XPATH_OBJ_TO_NODE(xpath_obj, 0);
   
    /* create encryption context, we don't need keys manager here */
    enc_ctx = xmlSecEncCtxCreate(mngr);
    if(!enc_ctx) {
        /*log*/
        goto done;
    }
    
    /* decrypt the data */
    if((xmlSecEncCtxDecrypt(enc_ctx, node) < 0) || (enc_ctx->result == NULL)) {
        /*log*/
        goto done;
    }
 
    if(enc_ctx->resultReplaced == 0) {
        /*log -- binary data*/
        goto done;
    }

#ifdef CSPACE_DEBUG
        /*should log properly*/
        xmlDocDump(stderr, doc);
#endif

    res = SUCC;

done:
    if (key)
        xmlSecKeyDestroy(key);

    if (xpath_obj)
        xmlXPathFreeObject(xpath_obj);

    if (mngr)
        xmlSecKeysMngrDestroy(mngr);

    if(enc_ctx) 
        xmlSecEncCtxDestroy(enc_ctx);

    return res;
}


static int register_namespaces(xmlXPathContext *xpath_ctx)
{
    if (xmlXPathRegisterNs(xpath_ctx, BAD_CAST SAML_PFX, BAD_CAST  SAML_HREF)) {
        return FAIL;
    }

    if (xmlXPathRegisterNs(xpath_ctx, BAD_CAST DSIG_PFX, BAD_CAST DSIG_HREF)) {
        return FAIL;
    }

    if (xmlXPathRegisterNs(xpath_ctx, BAD_CAST ENC_PFX, BAD_CAST ENC_HREF)) {
        return FAIL;
    }

    return SUCC;
}

static void logger_def(const char *msg, pc_log_level_t level, void *cb_ctx)
{
    printf("%s\n", msg);
}

static void *malloc_def(size_t size, void *cb_ctx)
{
    return malloc(size);
}

static void free_def(void *ptr, void *cb_ctx)
{
    free(ptr);
}

int cspace_process_context_set_key_file(process_context_t *ctx,
                                        const char* key_file)
{
    ctx->key_file = pc_malloc(strlen(key_file) + 1, ctx);
    if (!ctx->key_file)
        return FAIL;
    
    cspace_strcpy(ctx->key_file, key_file);
    return SUCC;
}

int cspace_process_context_set_ca_file(process_context_t *ctx,
                                        const char* ca_file)
{
    ctx->ca_file = pc_malloc(strlen(ca_file) + 1, ctx);
    if (!ctx->ca_file)
        return FAIL;
    
    cspace_strcpy(ctx->ca_file, ca_file);
    return SUCC;
}
void cspace_process_context_set_header_callback(process_context_t *ctx,
                                                set_header_cb_t set_header_fn,
                                                void *container)
{
    ctx->header_container = container;
    ctx->set_header_fn = set_header_fn;
}

process_context_t *cspace_process_context_create_default()
{
    return cspace_process_context_create_with_allocator(malloc_def, NULL,
                                                        free_def, NULL);
}

process_context_t *cspace_process_context_create_with_allocator(
                                        malloc_cb_t malloc_fn, void *mctx, 
                                        free_cb_t free_fn, void *fctx)
{
    process_context_t *ctx = NULL;
    ctx = (process_context_t *)malloc_fn(sizeof(process_context_t), mctx);
    if (!ctx)
        return NULL;
    
    ctx->allocator = (allocator_t *)malloc_fn(sizeof(allocator_t), mctx);
    if (!ctx->allocator) {
        free_fn(ctx, fctx);
        return NULL;
    }

    ctx->allocator->malloc_fn = malloc_fn;
    ctx->allocator->mctx = mctx;
    ctx->allocator->free_fn = free_fn;
    ctx->allocator->fctx = fctx;
   
    /* Set the default logger (stdio)
     * "User" can always change this afterwards */
    ctx->logger_fn = logger_def;
    ctx->logger_cb_ctx = NULL;

    ctx->key_file = NULL;
    ctx->ca_file = NULL;

    /* This call needs to be moved to user
     * For example in module we need this before the
     * ctx call*/
    /*if (!cspace_process_context_init()) {
        cspace_process_context_free(ctx);
        return NULL;
    }*/

    return ctx;
}

void cspace_process_context_free(process_context_t *ctx)
{
    /* we cannot free the header_container, mctx, or fctx here 
     * it should be done by the "user" */

    free_cb_t free_fn;
    void *fctx = NULL;

    if (!ctx)
        return;

    if (ctx->key_file)
        pc_free(ctx->key_file, ctx);
    
    fctx = ctx->allocator->fctx;
    free_fn = ctx->allocator->free_fn;
    if (ctx->allocator)
        pc_free(ctx->allocator, ctx);
    
    free_fn(ctx, fctx);
}

int cspace_process_context_init()
{
    xmlInitParser();
    LIBXML_TEST_VERSION

#if 0
    /*DTDs not used*/
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
    xmlSubstituteEntitiesDefault(1);
#ifndef XMLSEC_NO_XSLT
        xmlIndentTreeOutput = 1; 
#endif /* XMLSEC_NO_XSLT */
#endif

    /* Init xmlsec library */
    if(xmlSecInit() < 0) {
        fprintf(stderr, "Error: xmlsec initialization failed.\n");
        return FAIL;
    }

    /* Check loaded library version */
    if(xmlSecCheckVersion() != 1) {
        fprintf(stderr, 
                "Error: loaded xmlsec library version is not compatible.\n");
        return FAIL;
    }

    /* Load default crypto engine if we are supporting dynamic
     * loading for xmlsec-crypto libraries. Use the crypto library
     * name ("openssl", "nss", etc.) to load corresponding 
     * xmlsec-crypto library.
     */
#ifdef XMLSEC_CRYPTO_DYNAMIC_LOADING
    if(xmlSecCryptoDLLoadLibrary(BAD_CAST XMLSEC_CRYPTO) < 0) {
        fprintf(stderr, 
                "Error: unable to load default xmlsec-crypto library. "
                "Make sure\nthat you have it installed and check shared "
                "libraries path\n(LD_LIBRARY_PATH) envornment variable.\n");
        return FAIL; 
    }
#endif /* XMLSEC_CRYPTO_DYNAMIC_LOADING */

    /* Init crypto library */
    if(xmlSecCryptoAppInit(NULL) < 0) {
        fprintf(stderr, "Error: crypto initialization failed.\n");
        return FAIL;
    }

    /* Init xmlsec-crypto library */
    if(xmlSecCryptoInit() < 0) {
        fprintf(stderr, "Error: xmlsec-crypto initialization failed.\n");
        return FAIL;
    }

    return SUCC;
}

void cspace_process_context_halt()
{
    /* Shutdown xmlsec-crypto library */
    xmlSecCryptoShutdown();
    
    /* Shutdown crypto library */
    xmlSecCryptoAppShutdown();
    
    /* Shutdown xmlsec library */
    xmlSecShutdown();

    xmlCleanupParser();
}
